{{
Proramm         : VGA_640.spin
Info            : VT100 Terminal for AVR CP/M
Autor           : Joe G.
Version         : 01
Subversion      : 03
Funktion        : VT100 code interpreter
Komponenten     : VGA_HiRes_Color

Log

10.04.2013      Start, CR, LF
12.04.2013      code CA, BS
17.04.2013      fast insert line
26.04.2013      arrow keys, cursor (not) visible
27.04.2013      bug in ESC[K
25.05.2013      bell
03.01.2014      with German letters Ö,Ä,Ü,ö,ä,ü, bug in ESC[M
19.01.2014      second character font
28.01.2014      new VGA Mode
}}


CON
  cols     = 80                 'number of screen columns
  rows     = 30                 'number of screen rows
  chars    = rows*cols          'number of screen characters

  #0,  C_NOOP
  #9,  C_HT       'Moves cursor to next tab stop, or to right margin if there are no more tab stops.
  #10, C_IND      'Linefeed.
  #13, C_CR       'Moves cursor to left margin on current line.
  #14, C_G1       'Selects G1 character set designated by a select character set sequence.
  #15, C_G0       'Selects G0 character set designated by a select character set sequence.
  #31
  C_SGR           'Set character attributes
  C_DECSTBM       'Set scrolling region; parameters are top and bottom row.
  C_CUU           'Cursor up
  C_CUD           'Cursor down
  C_CUF           'Cursor right
  C_CUB           'Cursor left
  C_CUP           'Cursor Position
  C_HVP           'Horizontal/Vertical Position (same as CUP)
  C_RI            'Reverse linefeed. (Curser up)
  C_NEL           'Newline.
  C_DECSC         'Save current state (cursor coordinates, attributes, character sets pointed at by G0, G1).
  C_DECRC         'Restore state most recently saved by ESC 7.
  C_HTS           'Set tab stop at current column.
  C_TBC           'Clear TAB Stop. Without parameter: at current position.
  C_EL            'Erase in line
  C_ED            'Erase Display
  C_DCH           'Delete character(s)
  C_IL            'Insert line(s)
  C_DL            'Delete line(s)
  C_DSR           'Device status report: Answer is ESC [ 0 n (Terminal OK).
  C_DA            'Device Attributes (Answer ESC [ ? 6 c: "I am a VT102".)
  C_DECID         'DEC private identification. The terminal returns  ESC [ ? 6 c, claiming that it is a VT102.
  C_RIS           'Reset to Initial State (Power Up)
  C_DECALN        'DEC screen alignment test - fill screen with E's.
  'VT220
  C_ICH           'Insert Character(s) (VT220)
  C_ECH           'Erase character(s) (VT220)
  '
  'Modes:
  '
  C_IRM           'Insert/replace
'  C_DECIM         'Set insert mode. (default off)
  C_SRM           'Send/receive
  C_LNM           'Automatically follow echo of LF, VT or FF with CR. (default off)
  C_DECCOLM       'Number of columns (default off = 80 cols)
  C_DECSCNM       'Set reverse-video mode. (default off)
  C_DECOM         'Origin Mode; curser pos. in scrolling region. (default off)
  C_DECAWM[2]     'Set autowrap on/off. (default on)
  ' VT220
  C_DECTECM[2]    'Cursor not visible/visible

  C_DECCRM        'Display control chars. (default off)
  C_DECRST        'Reset Mode

OBJ
  vga : "VGA_HiRes_Text_Color"

VAR
  word  screen[chars]           ' screen character buffer
  byte  cursor[6]               ' cursor info array (see CursorPtr description above)
  long  sync, loc, xloc, yloc   ' sync used by VGA routine, others are local screen pointers

  long  char_color              ' foreground colour bits 3..0, background colour bits 7..4
  long  bg_fill
  byte  stat_wrap               'true if autowrap

  long  save_xloc
  long  save_yloc
  long  save_char_color         'TODO: init on power up/terminal reset


PUB start(BasePin)
''start vga

  vga.start(BasePin, @screen, @cursor, @sync)
  waitcnt(clkfreq * 1 + cnt)    'wait 1 second for cogs to start

''init cursor attributes
  cursor[0] := 127
  cursor[1] := 63
  cursor[2] := %110             'init cursor 0 to underscore with slow blink
  cursor[3] := 127
  cursor[4] := 63
  cursor[5] := 4


PUB bin(value, digits)
'' Print a binary number, specify number of digits

  repeat while digits > 32
    print("0")
    digits--

  value <<= 32 - digits

  repeat digits
    print((value <-= 1) & 1 + "0")

PUB dec(value) | i
'' Print a decimal number

  if value < 0
    -value
    print("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      print(value/i + "0")
      value //= i
      result~~
    elseif result or i == 1
      print("0")
    i /= 10

PUB hex(value, digits)
'' Print a hexadecimal number, specify number of digits

  repeat while digits > 8
    print("0")
    digits--

  value <<= (8 - digits) << 2

  repeat digits
    print(lookupz((value <-= 4) & $f : "0".."9", "A".."F"))


PUB binFP(value) | bitnum, bit, bitval
'' Prints FP long in special Binary format: sign, exp, mantissa

  repeat bitnum from 31 to 0
    bit := 1 << bitnum                ' create mask bit
    bitval := (bit & value) >> bitnum  ' extract bit and shift back to bit 0

    bin(bitval, 1)                    ' display one bit

    case bitnum
      27,20,16,12,8,4: print($20)       ' space after every 4 in group
      31,23: str(string("  "))        ' two after sign and exponent

PUB str(string_ptr)

'' Print a zero terminated string

  repeat while result := byte[string_ptr++]
    if result < " "
      cmd(result)
    else
      print(result)

PUB decaln

'' DECALN – Screen Alignment Display (DEC Private)
'' Fill the entire screen area with uppercase 'E's for screen focus and alignment.

  longfill(@screen, char_color + $00450045, chars/2)


PUB ed (what)
'' Erase in Display

  case what
    0:                                          'Erase from cursor to end of screen, inclusive.
      wordfill(@screen+loc<<1, bg_fill, chars-loc)
    1:                                          'Erase from beginning of screen to cursor, inclusive.
      wordfill(@screen, bg_fill, loc+1)
    2:                                          'Erase complete display. Cursor does not move.
      longfill(@screen, bg_fill, chars/2)



PUB el (what)
'' Erase in line

  case what
    0:                                  'Erase from cursor to end of line, inclusive.
      wordfill(@screen+loc*2, bg_fill, cols-xloc)
    1:                                  'Erase from beginning of line to cursor, inclusive.
      wordfill(@screen+yloc*cols*2, bg_fill, xloc+1)
    2:                                  'Erase complete line. Cursor does not move.
      wordfill(@screen+yloc*cols*2, bg_fill, cols)


PUB ech (count)
'' Erase count characters on current line.

  wordfill(@screen+loc*2, bg_fill, count #> 1 <# cols-xloc)


PUB dch (count)
'' Delete count characters on current line.

  count := count #> 1 <# cols-xloc

  wordmove(@screen+loc*2, @screen+(loc+count)*2, cols-(xloc+count))

  'VT102 spec:
  'result := screen[yloc*cols + (cols-1)] & $FF00 | $20   'space with color from last char in line
  'VT220 spec:
  result := bg_fill                                 'TODO: set colors to normal!
  wordfill(@screen+(loc+cols-(xloc+count))*2, result,  count)


PUB ich (count)
'' Insert count blank characters at the cursor position,
'' with the character attributes set to normal. (VT220)

  count := count #> 1 <# cols-xloc

  wordmove(@screen+(loc+count)*2, @screen+loc*2, cols-(xloc+count))

  'result := char_color + $20                          'TODO: set colors to normal!
  wordfill(@screen+loc*2, bg_fill,  count)


PUB color(ColorVal)
''reset screen colors

  char_color := ColorVal << 8 + ColorVal << 24
  bg_fill := char_color + $00200020


PUB cursloc(x, y)
'' move cursor to x, y position

  xloc := cursor[0] := x <# cols-1
  yloc := cursor[1] := y <# rows-1
  loc  := xloc + yloc*cols

PUB get_cursor
'' return cursor position (0 based)

  return yloc<<16 + xloc


PRI scroll_up | p, len
        p := @screen + cols * 2         'start of 2nd row
        len := (chars - cols)/2
        longmove(@screen, p, len)    'shift screen up one line

        'clear last line
        p := @screen + (rows-1)*cols * 2    'row start position in pointer
        longfill(p, bg_fill, cols/2)      'fill line with space

PRI scroll_down | p, len
        p := @screen + cols * 2                 'start of row 2
        len := (chars - cols)/2
        longmove(p, @screen, len)      'shift screen down one line

        'clear top line
        wordfill(@screen, bg_fill, cols) 'fill line with space


PUB print(c)
'' Print a character

  screen[loc++] := char_color  + c 'output the character
'  screen[yloc*cols+xloc] := char_color  + c 'output the character

  if xloc == cols-1         'bump the xloc
    if stat_wrap
      xloc := 0
      if yloc == rows-1
        scroll_up
        loc -= cols
      else
        cursor[1] := ++yloc
    else
      --loc
  else
    cursor[0] := ++xloc


PUB cmd(c) | i, j, k, len
'' execute a control function

  case c

    C_HT:                        ' tab command
      cursloc((xloc+8) & constant(!7), yloc) 'tabstop every 8 cols for now. (TODO)

    C_IND:                       'Index/LF
      if yloc == rows-1          'last row?
        scroll_up               'cursor position not changed
      else                       'rows 0-29
        yloc++
        cursor[1] := yloc       'new cursor position
        loc += cols             'next output position

    C_CR:                        'CR - move to 1st char of line
      xloc := cursor[0] := 0
      loc := xloc + yloc*cols

    C_CUU:                      'Cursor up
      if yloc                   'skip if at top of screen
        cursor[1] := --yloc     'reset 'y' cursor position
        loc -= cols             'move screen entry index back one row

    C_CUD:                      'Cursor down
      if yloc <> rows - 1       'skip if at bottom of screen
        cursor[1] := ++yloc
        loc += cols

    C_CUF:                      'Cursor right
      if xloc <> cols - 1       'skip if at right margin of screen
        cursor[0] := ++xloc
        ++loc

    C_CUB:                       'Cursor left, BS
      if xloc > 0
        loc--                    'screenpointer - 1
        xloc--                   'cursorpointer - 1
        cursor[0] := xloc        'set new cursor position

    C_RI:                        'Reverse LF
      if yloc == 0               'on first row?
        scroll_down             'cursor position not changed
      else                       'rows 0-29
        cursor[1] := --yloc     'new row cursor position
        loc -= cols             'next output position

    C_DECSC:                        'save cursor
      save_xloc := xloc
      save_yloc := yloc
      save_char_color := char_color

    C_DECRC:                        'restore cursor
      cursor[0] := xloc := save_xloc
      cursor[1] := yloc := save_yloc
      loc := xloc + yloc*cols
      char_color := save_char_color

    C_IL:                        'insert line and roll down "ESC[L"
      yloc := cursor[1]         'row position
      i := @screen              'screen addr

      if yloc < rows-1          'rows 0-29
        j := i + yloc*cols*2    'yloc addr, current position
        k := j + cols*2         'next line
        len := (chars - (yloc+1)*cols)>>1
        longmove(k, j, len)     'shift screen down
        wordfill(j, bg_fill, cols)  'fill line with space

      if yloc == rows-1         'rows 29, last row
        j := i + yloc*cols*2
        wordfill(j, bg_fill, cols)  'delete line

    C_DL:                        'delete line and roll up "ESC[M"
      yloc := cursor[1]         'row position
      i := @screen              'screen address

      if yloc < rows-1          'rows 0-29
        j := i + yloc*cols*2    'yloc addr, current position
        k := j + cols*2         'next line
        len := (chars - (yloc+1)*cols)>>1
        longmove(j, k, len)     'shift screen up
        j := i + (rows-1)*cols*2'last line
        wordfill(j, bg_fill, cols)  'delete last line

      if yloc == rows-1         'rows 29, last row
        j := i + yloc*cols*2
        wordfill(j, bg_fill, cols)  'delete last line

    C_DECTECM+0:                 'TODO
      cursor[2] := %100         'cursor 0 not visible

    C_DECTECM+1:                 'TODO
      cursor[2] := %110         'cursor 0 visible and underscore with slow blink

    C_DECAWM+0:  'TODO           'Autowrap off
      stat_wrap := FALSE

    C_DECAWM+1:  'TODO           'Autowrap on
      stat_wrap := TRUE
