'' PropCOMM VT100_VGA V2.0b Original By Jeff Ledger

{{
Info            : VT100 Terminal for AVR CP/M
Autor           : Joe G.
Version         : 1.3
Subversion      : 02
Funktion        : Keyboard Driver DE, VGA-640x480 Driver, V24 Driver, VT100 Terminal
Komponenten     : VGA_640
                  Keyb-DE
                  FullDuplexSerial
                  VGA_HiRes_Text_Color

Log

25.03.2013      Start
                V24 Buffer 256 Byte
                Key-DE
26.03.2013      ESC Code Terminal Reset
10.04.2013      new codes
13.04.2013      BS
14.04.2013      ESC [ Y;X H
26.04.2013      arrow Keys, cursor (not) visible, colormanagement
25.05.2013      Bell, CP/M Reset
26.05.2013      all VT100 colors, hot key Alt+i (toggle inverse)
02.01.2014      makes Cursor visible by Reset (F1 und ESC c), correct storage location in the Rx&Tx buffer
03.01.2014      with German letters Ä,Ö,Ü,ä,ö,ü,ß and ESC[H, ESC[f, bug in ESC[M
19.01.2014      second character font
28.01.2014      new VGA Mode 640x480 full color




# Current VT-100 Code list

## ESC sequences:

ESC c     RIS      Reset.
ESC D     IND      Linefeed.
ESC E     NEL      Newline.
ESC H     HTS      Set tab stop at current column.
ESC M     RI       Reverse linefeed.
ESC Z     DECID    DEC private identification. The terminal returns the
                   string  ESC [ ? 6 c, claiming that it is a VT102.
ESC 7     DECSC    Save current state (cursor coordinates,
                   attributes, character sets pointed at by G0, G1).
ESC 8     DECRC    Restore state most recently saved by ESC 7.
ESC [     CSI      Control sequence introducer
ESC # 8   DECALN   DEC screen alignment test - fill screen with E's.
ESC (              Start sequence defining G0 character set
ESC ( B               Select default (ISO 8859-1 mapping)
ESC ( 0               Select VT100 graphics mapping
ESC )              Start sequence defining G1
                   (followed by one of B, 0 as above).
ESC >     DECPNM   Set numeric keypad mode
ESC =     DECPAM   Set application keypad mode


## ECMA-48 CSI sequences

CSI (or ESC [) is followed by a sequence of parameters, at most NPAR
(16), that are decimal numbers separated by semicolons.  An empty or
absent parameter is taken to be 0.  The sequence of parameters may be
preceded by a single question mark.

The action of a CSI sequence is determined by its final character.

@   ICH       Insert the indicated # of blank characters.
A   CUU       Move cursor up the indicated # of rows.
B   CUD       Move cursor down the indicated # of rows.
C   CUF       Move cursor right the indicated # of columns.
D   CUB       Move cursor left the indicated # of columns.
H   CUP       Move cursor to the indicated row, column (origin at 1,1).
J   ED        Erase display (default: from cursor to end of display).
              ESC [ 1 J: erase from start to cursor.
              ESC [ 2 J: erase whole display.
K   EL        Erase line (default: from cursor to end of line).
              ESC [ 1 K: erase from start of line to cursor.
              ESC [ 2 K: erase whole line.
L   IL        Insert the indicated # of blank lines.
M   DL        Delete the indicated # of lines.
P   DCH       Delete the indicated # of characters on current line.
X   ECH       Erase the indicated # of characters on current line.
c   DA        Answer ESC [ ? 6 c: "I am a VT102".
f   HVP       Move cursor to the indicated row, column.
g   TBC       Without parameter: clear tab stop at current position.
              ESC [ 3 g: delete all tab stops.
h   SM        Set Mode (see below).
l   RM        Reset Mode (see below).
m   SGR       Set attributes (see below).
n   DSR       Status report (see below).
q   DECLL     Set keyboard LEDs.
              ESC [ 0 q: clear all LEDs
              ESC [ 1 q: set Scroll Lock LED
              ESC [ 2 q: set Num Lock LED
              ESC [ 3 q: set Caps Lock LED
r   DECSTBM   Set scrolling region; parameters are top and bottom row.



## ECMA-48 Set Graphics Rendition

The ECMA-48 SGR sequence ESC [ parameters m sets display attributes.
Several attributes can be set in the same sequence, separated by
semicolons.  An empty parameter (between semicolons or string
initiator or terminator) is interpreted as a zero.

param   result
0       reset all attributes to their defaults
1       set bold
2       set half-bright (simulated with color on a color display)
4       set underscore (simulated with color on a color display)
        (the colors used to simulate dim or underline are set
        using ESC ] ...)
5       set blink
7       set reverse video
21      set normal intensity (ECMA-48 says "doubly underlined")
22      set normal intensity
24      underline off
25      blink off
27      reverse video off
30      set black foreground
31      set red foreground
32      set green foreground
33      set brown (yellow) foreground
34      set blue foreground
35      set magenta foreground
36      set cyan foreground
37      set white foreground
38      set underscore on, set default foreground color
39      set underscore off, set default foreground color
40      set black background
41      set red background
42      set green background
43      set brown (yellow) background
44      set blue background
45      set magenta background
46      set cyan background
47      set white background
49      set default background color


## ECMA-48 Mode Switches

Control sequence 'ESC [ Ps ; ... h' sets one or more modes specified
by selective parameters (Ps) in the parameter string.
Control sequence 'ESC [ Ps ; ... l' resets one or more modes specified
by selective parameters (Ps) in the parameter string.

Ps   Mnemonic   Description

3    DECCRM     (default off): Display control chars.
4    DECIM      (default off): Set insert mode.
20   LF/NL      (default off): Automatically follow echo of LF, VT or FF
                with CR.

## DEC Private Mode (DECSET/DECRST) sequences

'ESC [ ? Ps ; ... h' sets one or more private modes.
'ESC [ ? Ps ; ... l' resets one or more private modes.

Ps   Mnemonic   Description
3    DECCOLM    (default off = 80 columns): 80/132 col mode switch.
                NOTE: When you change the # of columns per line, the screen
                is erased! This also sets the scrolling region for full screen.
5    DECSCNM    (default off): Set reverse-video mode.
6    DECOM      (default off): When set, cursor addressing is relative
                to the upper left corner of the scrolling region.
7    DECAWM     (default on): Set autowrap on.  In this mode, a graphic
                character emitted after column 80 (or column 132 of DECCOLM is
                on) forces a wrap to the beginning of the following line
                first.
25   DECTECM    (default on): Make cursor visible.


## ECMA-48 Status Report Commands

ESC [ 5 n       Device status report (DSR):
                Answer is ESC [ 0 n (Terminal OK).
                       or ESC [ 3 n (Terminal not OK)
ESC [ 6 n       Cursor position report (CPR): Answer is ESC [ y ; x R, where
                x,y is the cursor location.

## Linux Console Private CSI Sequences
(not implemented yet)

The following sequences are neither ECMA-48 nor native VT102.  They
are native to the Linux console driver.  Colors are in SGR
parameters: 0 = black, 1 = red, 2 = green, 3 = brown, 4 = blue, 5 =
magenta, 6 = cyan, 7 = white.

ESC [ 1 ; n ]       Set color n as the underline color
ESC [ 2 ; n ]       Set color n as the dim color
ESC [ 8 ]           Make the current color pair the default attributes.


}}

{{
Terminal Input Parser
see http://www.vt100.net/emu/dec_ansi_parser


State           Event           Next State      Action
--------------------------------------------------------------
'anywhere'      00-17,19,1C-1F  -               execute
                18, 1A          ground          execute
                1B              escape          clear_esc
--------------------------------------------------------------
ground          20-7F           -               print
--------------------------------------------------------------
escape          7F              -               ignore
                20-2F           esc_interm      collect
                5B              csi_entry       clear_csi
                30-5A,5C-7E     ground          esc_dispatch
--------------------------------------------------------------
esc_interm      7F              -               ignore
                20-2F           -               collect
                30-7E           ground          esc_dispatch
--------------------------------------------------------------
csi_entry       7F              -               ignore
                20-2F           csi_interm      collect
                3A              csi_ignore      -
                30-39,3B        csi_param       csi_param
                3C-3F           csi_param       collect
                40-7E           ground          csi_dispatch
--------------------------------------------------------------
csi_interm      7F              -               ignore
                20-2F           -               collect
                30-3F           csi_ignore      -
                40-7E           ground          csi_dispatch
--------------------------------------------------------------
csi_ignore      20-3F,7F        -               ignore
                40-7E           ground          -
--------------------------------------------------------------
csi_param       7F              -               ignore
                20-2F           csi_interm      collect
                30-39,3B        -               csi_param
                3A,3C-3F        csi_ignore      -
                40-7E           ground          csi_dispatch
--------------------------------------------------------------

}}

CON
  'clock configuration
  _clkmode  = xtal1 + pll16x
  _xinfreq  = 5_000_000         '5 MHz Quarz AVR CP/M

  'pin configuration
  key_data  = 26                'Keyboard Data Pin
  key_clk   = 27                'Keyboard CLK Pin
  video     = 16                'Video output Pin
  serXmit   = 30                'Serial Transmit
  serRecv   = 31                'Serial Receive

  bell     = 1                  'Beep
  CP_M     = 0                  'CP/M Reset


  'seriell port
  baud      = 115200           'Default AVR CP/M is 115200
  'baud      = 9600

'                      DRGB
  black             = %0000     ' black
  dblack            = %1000
  dred              = %0100     ' dark red
  dgreen            = %0010     ' dark green
  dblue             = %0001     ' dark blue
  dyellow           = %0110     ' dark yellow
  dmagenta          = %0101     ' dark magenta
  dcyan             = %0011     ' dark cyan
  gray              = %0111     ' gray
  red               = %1100     ' red
  green             = %1010     ' green
  blue              = %1001     ' blue
  yellow            = %1110     ' yellow
  magenta           = %1101     ' magenta
  cyan              = %1011     ' cyan
  white             = %1111     ' white

'default color
  fg_color          = dgreen    'foreground color  (default dgreen)
  bg_color          = black     'background color  (default black)

'Control Characters Recognized by VT102

  NUL           = $00           'Ignored when received (not stored in input buffer) and used as a fill character.
  ETX           = $03           'Can be selected as a half-duplex turnaround character.
  EOT           = $04           'Can be selected as a disconnect character or half-duplex turnaround character. When used as a turnaround character, the disconnect character is DLE-EOT.
  ENQ           = $05           'Transmits answerback message.
  BEL           = $07           'Generates bell tone.
  BS            = $08           'Moves cursor to the left one character position; if cursor is at left margin, no action occurs.
  HT            = $09           'Moves cursor to next tab stop, or to right margin if there are no more tab stops.
  LF            = $0A           'Causes a linefeed or a new line operation. (See Linefeed/New Line). Also causes printing if auto print operation is selected.
  VT            = $0B           'Processed as LF.
  FF            = $0C           'Processed as LF. FF can also be selected as a half-duplex turnaround character.
  CR            = $0D           'Moves cursor to left margin on current line. CR can also be selected as a half-duplex turnaround character.
  SO            = $0E           'Selects G1 character set designated by a select character set sequence.
  SI            = $0F           'Selects G0 character set designated by a select character set sequence.
  DC1           = $11           'Processed as XON. DC1 causes terminal to continue transmitting characters.
  DC3           = $13           'Processed as XOFF. DC3 causes terminal to stop transmitting all characters except XOFF and XON. DC3 can also be selected as a half-duplex turnaround character.
  CAN           = $18           'If received during an escape or control sequence, cancels the sequence and displays substitution character ([]).
  SUBST         = $1A           'Processed as CAN.
  ESC           = $1B           'Processed as a sequence introducer.
  DEL           = $7F

  bspKey            = $C8
  delKey            = $BA
  ArrowTop          = $C2
  ArrowDown         = $C3
  ArrowLeft         = $C0
  ArrowRight        = $C1

  escape            = $CB
  backspace         = 8
  fReturn           = 13
  fLinefeed         = 10
  fEof              = -1
  CSI_PAR_MAX        = 10        'Maximum number of escape sequence parameters

  ' Input parser states
  #0, S_GROUND 'comment
  S_ESCAPE
  S_ESCAPE_INTERMEDIATE
  S_CSI_ENTRY
  S_CSI_INTERMEDIATE
  S_CSI_IGNORE
  S_CSI_PARAM

Var
  byte  color

  ' Input parser state variables
  long   intermediate_char
  long   csi_par_cnt
  long   csi_par[CSI_PAR_MAX]

OBJ
  text   : "VGA_640"                   ' Full Color VGA-640x480 Driver
  key    : "keyb-DE"                   ' new Keyboard Driver (german)
  ser    : "FullDuplexSerial"          ' Full Duplex Serial Controller

PUB start | c                                           'c only for debuging

  dira[bell] := 1               'output
  dira[CP_M] := 0               'input
  text.start(video)                                     'start terminal screen
  key.start(key_data,key_clk)                           'start new keyboard driver
  ser.start(serRecv,serXmit,0,baud)                     'start V24 to AVR CP/M
  vt_reset                                              '(pre)set parameters
  titlescreen                                           'print startmessage

{
  repeat
    c := key.getkey
    Text.hex(c,3)
    Text.str(string(" : "))
    text.print(c)
    text.print($0d)
    ser.tx(c)

}
    terminal                                            'start VT100 Terminal Programm


PRI titlescreen
  beep
  text.str(string($0D,"AVR CP/M VT100 Terminal V1.3 lc 7",$0D,$0A))
  text.dec(baud)
  text.str(string(",8,N,1",$0D,$0A))
  text.str(string("Terminal Ready",$0D,$0A))

PRI vt_reset
  text.cmd(text#C_DECAWM+1)                         'Autowrap on
  text.cmd(text#C_DECTECM+1)                        'makes cursor visible
  color := bg_color << 4 + fg_color                 'default color
  text.color(color)
  text.ed(2)
  text.cursloc(0,0)

PRI beep

  outa[bell] := 1        'set bell high
  waitcnt(clkfreq/8 + cnt)'wait a moment
  outa[bell] := 0        'set bell low

PRI CP_M_Reset
  dira[CP_M] := 1                                       'output
  outa[CP_M] := 0        'set CP/M reset low
  waitcnt(clkfreq/128 + cnt)'wait a moment
  outa[CP_M] := 1        'set CP/M reset high
  dira[CP_M] := 0                                       'input


PRI terminal | ch, vt100

   vt100 := S_GROUND

   repeat
      ch := key.key

      if ch > 0

         if ch > 608 and ch < 635                 'CRTL code
            ch:=ch-608                            'redefine as CTRL+char

         if ch > 451 and ch < 477                 'Ä,Ü,Ö
            ch:=ch-256                            'redefine as ASCII Letter Ä,Ü,Ö

         if ch == 208                                'if F1
            ch := 0
            vt_reset
            titlescreen

         if ch == 217                                'if F10 (old F12)
            CP_M_Reset                               'CP/M Reset

         if ch == 1129                               'Alt + i
            ch := 0
            color := (0 + bg_color) << 4 + fg_color     'default color
            text.color(color)
            beep                                       'beep

         case ch
          bspKey     : ser.tx($08)                      'send BS

          delKey     : ser.tx($7F)                      'send DEL

          ArrowTop   : ser.tx($1B)                      'send ESC [ A
                       ser.tx($5B)
                       ser.tx($41)

          ArrowDown  : ser.tx($1B)                      'send ESC [ B
                       ser.tx($5B)
                       ser.tx($42)

          ArrowRight : ser.tx($1B)                      'send ESC [ C
                       ser.tx($5B)
                       ser.tx($43)

          ArrowLeft  : ser.tx($1B)                      'send ESC [ D
                       ser.tx($5B)
                       ser.tx($44)

          0:                                             'send nothing

          OTHER      : ser.tx(ch)                    'send code

'      if (ch := ser.rxcheck) > 0               'receive char, ignore NUL
      ch := ser.rxcheck                         'receive char, ignore NUL

        if ch < $20
          'Process control characters immediately. (everywhere pseudostate)
          case ch

            'ENQ:                               'ENQ: Transmits answerback message.
                'TODO
            BEL:                                'BEL: Generates bell tone.
                beep
            BS:                                 'BS: Moves cursor to the left one character position; if cursor is at left margin, no action occurs.
                text.cmd(text#C_CUB)
                                                'HT: Moves cursor to next tab stop, or to right margin if there are no more tab stops.
            HT:
                text.cmd(text#C_HT)             'CR: Moves cursor to left margin on current line.
            CR:
                text.cmd(text#C_CR)
                                                'LF: Causes a linefeed or a new line operation. (See Linefeed/New Line).
                                                'VT: Processed as LF.
                                                'FF: Processed as LF.
            LF, VT, FF:
                text.cmd(text#C_IND)

            'SO                                 'Selects G1 character set designated by a select character set sequence.
            'SI                                 'Selects G0 character set designated by a select character set sequence.
            'DC1                                'Processed as XON. DC1 causes terminal to continue transmitting characters.
            'DC3                                'Processed as XOFF. DC3 causes terminal to stop transmitting all characters except XOFF and XON. DC3 can also be selected as a half-duplex turnaround character.
                'TODO

                                                'CAN: If received during an escape or control sequence, cancels the sequence and displays substitution character ([]).
                                                'SUB: Processed as CAN.
            CAN, SUBST:
                'TODO: text.print(**subst char**)
                vt100 := S_GROUND

            ESC:                                'ESC: Processed as a sequence introducer.
                'clear_esc
                intermediate_char := 0
                vt100 := S_ESCAPE

        else

          case vt100
            S_GROUND:
                text.print(ch)

            S_ESCAPE:
                case ch
                  DEL:
                      'ignore

                  $20..$2F:                      '" ".."/"
                      collect(ch)
                      vt100 := S_ESCAPE_INTERMEDIATE

                  $5B:                           '"["
                      'clear csi
                      csi_par_cnt := 0
                      csi_par[0] := 0
                      vt100 := S_CSI_ENTRY

                  other:
                      esc_dispatch(ch)
                      vt100 := S_GROUND

            S_ESCAPE_INTERMEDIATE:
                case ch
                  DEL:
                      'ignore

                  $20..$2F:                      '" ".."/"
                      collect(ch)
                      vt100 := S_CSI_INTERMEDIATE

                  other:
                      esc_dispatch(ch)
                      vt100 := S_GROUND

            S_CSI_ENTRY:
                case ch
                  DEL:
                      'ignore

                  $20..$2F:                      '" ".."/"
                      collect(ch)

                  $3A:                           '":"
                      vt100 := S_CSI_IGNORE

                  $30..$39,$3B:                  '"0".."9",";"
                      param(ch)
                      vt100 := S_CSI_PARAM

                  $3C..$3F:                      '"<".."?"
                      collect(ch)
                      vt100 := S_CSI_PARAM

                  other:
                      csi_dispatch(ch)
                      vt100 := S_GROUND

            S_CSI_INTERMEDIATE:
                case ch
                  DEL:
                      'ignore

                  $20..$2F:                      '" ".."/"
                      collect(ch)

                  $30..$3F:                      '"0".."?"
                      vt100 := S_CSI_IGNORE

                  other:
                      csi_dispatch(ch)
                      vt100 := S_GROUND

            S_CSI_IGNORE:
                case ch
                  $20..$3F,DEL:                  '" ".."?",DEL
                      'ignore

                  other:
                      vt100 := S_GROUND

            S_CSI_PARAM:
                case ch
                  DEL:
                      'ignore

                  $20..$2F:                      '" ".."/"
                      collect(ch)
                      vt100 := S_CSI_INTERMEDIATE

                  $3A,$3C..$3F:                  '":","<".."?"
                      vt100 := S_CSI_IGNORE

                  $30..$39,$3B:                  '"0".."9",";"
                      param(ch)

                  other:
                      csi_dispatch(ch)
                      vt100 := S_GROUND



PRI collect (ch)
   if intermediate_char == 0                    'Store only one (the first) intermediate char
      intermediate_char := ch

PRI param (ch)
   if csi_par_cnt < CSI_PAR_MAX
      if ch == ";"
         csi_par_cnt++
         if csi_par_cnt < CSI_PAR_MAX
            csi_par[csi_par_cnt] := 0
      else
         csi_par[csi_par_cnt] := csi_par[csi_par_cnt] * 10 + ch - "0"

PRI get_param (index)
   if index > csi_par_cnt
      return 0

   return csi_par[index]


PRI esc_dispatch (ch)

  case intermediate_char
    0:
      case ch
        "c":                                    'RIS    Reset to Initial State (Power Up)
          vt_reset
          titlescreen
        "D":                                    'IND    Linefeed.
          text.cmd(text#C_IND)
        "E":                                    'NEL    Newline.
          text.cmd(text#C_CR)
          text.cmd(text#C_IND)
        "H":                                    'HTS    Set tab stop at current column.
          'TODO?
        "M":                                    'RI     Reverse linefeed. (Curser up)
          text.cmd(text#C_RI)
        "Z":                                    'DECID  DEC private identification. The terminal returns  ESC [ ? 6 c, claiming that it is a VT102.
          tx_csi(string("?6c"))
        "7":                                    'DECSC  Save current state (cursor coordinates, attributes, character sets pointed at by G0, G1).
          text.cmd(text#C_DECSC)
        "8":                                    'DECRC  Restore state most recently saved by ESC 7.
          text.cmd(text#C_DECRC)

    "?":                                        'Private functions
      'none (yet)
    other:
      case ch
        "8":                                    'DECALN   DEC screen alignment test - fill screen with E's.
           if intermediate_char == "#"
              Text.decaln


PRI csi_dispatch (ch) | i

  case intermediate_char
    0:
      case ch
        "@":                                    'ICH    Insert character(s) (VT220)
          text.ich(get_param(0))
        "A":                                    'CUU    Move cursor up
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUU)

        "B":                                    'CUD    Move cursor down
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUD)

        "C":                                    'CUF    Move cursor right
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUF)

        "D":                                    'CUB    Move cursor left
          repeat (get_param(0) #>1)
            text.cmd(text#C_CUB)

        "H", "f":                               'CUP,HVP Move cursor to row, column (origin at 1,1).
          text.cursloc((get_param(1) #>1) - 1, (get_param(0) #>1) - 1)

        "J":                                    'ED     Erase Display
          text.ed(get_param(0))

        "K":                                    'EL     Erase in line
          text.el(get_param(0))

        "L":                                    'IL     Insert line(s)
          repeat (get_param(0) #>1)
            text.cmd(text#C_IL)

        "M":                                    'DL     Delete line(s)
          repeat (get_param(0) #>1)
            text.cmd(text#C_DL)

        "P":                                    'DCH    Delete character(s)
          text.dch(get_param(0))

        "X":                                    'ECH    Erase character(s) (VT220)
          text.ech(get_param(0))

        "m":                                    'SGR    Set character attributes
          repeat i from 0 to csi_par_cnt
            case result := get_param(i)
              0:                                      'All attributes off
                color := bg_color << 4 + fg_color     '(default color)
              1:                                      'bold (increased intensity)
                color |= %00001000                    'set bit for high color
              2:                                      'half-bright (simulated with color)
              4:                                      'underline (simulated with color)
              5:                                      'blink
              7:                                      'negative (reverse) image
                result := (color>>4) & 7 | (color&7)<<4
                color := (color & %10001000) | result
              22:                                     'bold off (normal intensity)
                color &= %11110111     'bold off
              24:                                     'not underlined
              25:                                     'not blinking
              27:                                     'positive image
              30..37:                                 'fg (character) color
                color := (color & %11111000) | ((result-30) >< 3)
              39:                                     'default fg color
                color := (color & %11111000) | fg_color
              40..47:                                 'bg color
                color := (((result-40) >< 3) << 4) | (color & %00001111)
              49:                                     'default bg color
                color := (bg_color << 4) | (color & %00001111)

          text.color(color)

        "c":                                    'DA     Answer ESC [ ? 6 c: "I am a VT102".
          tx_csi(string("?6c"))

        "g":                                    'TBC    Without parameter: clear tab stop at current position.
          'TODO or not TODO?                    '       ESC [ 3 g: delete all tab stops.

        "h":                                    'SM     'Set Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCRM Display control chars. (default off)
                ' TODO?
              4:                                'DECIM  Set insert mode. (default off)
                ' TODO?
              20:                               'LF/NL  Automatically follow echo of LF, VT or FF with CR. (default off)
                ' TODO:

        "l":                                    'RM     'Reset Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCRM Display control chars. (default off)
                ' TODO?
              4:                                'DECIM  Set insert mode. (default off)
                ' TODO?
              20:                               'LF/NL  Automatically follow echo of LF, VT or FF with CR. (default off)
                ' TODO:

        "r":                                    'DECSTBM Set scrolling region; parameters are top and bottom row.
          'TODO or not TODO?

        "n":                                    'DSR    Status report
           repeat i from 0 to csi_par_cnt
              case get_param(i)
                5:                              'DSR    Device status report: Answer is ESC [ 0 n (Terminal OK).
                  tx_csi(string("0n"))
                6:                              'CPR    Cursor position report: Answer is ESC [ y ; x R
                  'repeat 1
                  waitcnt(10*clkfreq/baud + cnt)          'debug wait 1 char time
                  result := text.get_cursor
                  ser.tx(ESC)
                  ser.tx("[")
                  ser.dec(result >> 16 + 1)
                  ser.tx(";")
                  ser.dec(result & $FF + 1)
                  ser.tx("R")
        "]":                                    'From Linux console
          case get_param(0)
            1:                                  'Set color n as the underline color
            2:                                  'Set color n as the dim color
            8:                                  'Make the current color pair the default attributes.

    "?":                                        'Private mode
      case ch
        "h":                                    'DECSET Set Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCOLM Number of columns (default off = 80 cols)
                text.ed(2)                      'Not supported, but clear the sceen anyway
                text.cursloc(0,0)
              5:                                'DECSCNM Set reverse-video mode. (default off)
              6:                                'DECOM   Origin Mode; curser pos. in scrolling region. (default off)
              7:                                'DECAWM  Set autowrap on/off. (default on)
                text.cmd(text#C_DECAWM+1)
              25:                               'DECTECM Cursor visible
                text.cmd(text#C_DECTECM+1)

        "l":                                    'DECRST Reset Mode
          repeat i from 0 to csi_par_cnt
            case get_param(i)
              3:                                'DECCOLM Number of columns (default off = 80 cols)
                text.ed(2)                      'Not supported, but clear the sceen anyway
                text.cursloc(0,0)
              5:                                'DECSCNM Set reverse-video mode. (default off)
              6:                                'DECOM   Origin Mode; curser pos. in scrolling region. (default off)
              7:                                'DECAWM  Set autowrap on/off. (default on)
                text.cmd(text#C_DECAWM+0)
              25:                               'DECTECM Cursor not visible
                text.cmd(text#C_DECTECM+0)

    other:
      '


PRI tx_csi(seq)

  'repeat 5
  waitcnt(10*clkfreq/baud + cnt)          'debug wait 1 char time
  ser.tx($1B)                             'send ESC [
  ser.tx($5B)
  ser.str(seq)

{{

┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
